@@ -600,6 +600,37 @@ pub type PackageDecl<'a> = Package<'a>;
 #[deprecated(note = "use ast::Modport instead")]
 pub type ModportDecl<'a> = Modport<'a>;
 
+/// A parsing result which may be ambiguous.
+#[derive(Debug, PartialEq, Eq, Clone)]
+pub enum Ambiguous<R> {
+    Unique(R),
+    Ambiguous(Vec<R>),
+}
+
+impl<'a, R> WalkVisitor<'a> for Ambiguous<R>
+where
+    R: WalkVisitor<'a>,
+{
+    fn walk(&'a self, visitor: &mut dyn Visitor<'a>) {
+        match self {
+            Self::Unique(x) => x.walk(visitor),
+            Self::Ambiguous(x) => x.walk(visitor),
+        }
+    }
+}
+
+impl<'a, R> ForEachNode<'a> for Ambiguous<R>
+where
+    R: ForEachNode<'a>,
+{
+    fn for_each_node(&'a self, each: &mut dyn FnMut(&'a dyn AnyNode<'a>)) {
+        match self {
+            Self::Unique(x) => x.for_each_node(each),
+            Self::Ambiguous(x) => x.for_each_node(each),
+        }
+    }
+}
+
 /// All things being compiled.
 #[moore_derive::node]
 #[indefinite("root")]
@@ -1514,12 +1545,15 @@ pub struct Identifier {
     pub name: Name,
 }
 
-#[moore_derive::visit]
+/// An argument to a a function or task call.
+#[moore_derive::node]
+#[indefinite("call argument")]
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct CallArg<'a> {
-    pub span: Span,
-    pub name_span: Span,
-    pub name: Option<Name>,
+    /// An optional name of the argument, e.g. `.foo(x)` instead of just `x`.
+    pub name: Option<Spanned<Name>>,
+    /// An optional value to assign to the argument. Omitting this causes the
+    /// call to use the default argument in the function/task declaration.
     pub expr: Option<Expr<'a>>,
 }
 
@@ -1692,6 +1726,7 @@ pub enum ConstraintItemData<'a> {
 #[indefinite("subroutine declaration")]
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct SubroutineDecl<'a> {
+    #[forward]
     pub prototype: SubroutinePrototype<'a>,
     pub items: Vec<SubroutineItem<'a>>,
 }
@@ -1701,11 +1736,17 @@ pub struct SubroutineDecl<'a> {
 #[indefinite("subroutine prototype")]
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct SubroutinePrototype<'a> {
+    /// Whether this is a function or a task.
     pub kind: SubroutineKind,
+    /// The default lifetime of the subroutine.
     pub lifetime: Option<Lifetime>,
+    /// The name of the subroutine.
     #[name]
     pub name: Spanned<Name>,
-    pub args: Vec<SubroutinePort<'a>>,
+    /// The arguments of the function or task. `None` if no parentheses have
+    /// been provided; `Some(vec![...])` otherwise.
+    pub args: Option<Vec<SubroutinePort<'a>>>,
+    /// The return type.
     pub retty: Option<Type<'a>>,
 }
 
@@ -1717,21 +1758,38 @@ pub enum SubroutineKind {
 }
 
 /// A function or task port.
+///
+/// This is the `inout logic [7:0]` part in `inout logic [7:0] x[3]`.
+///
+/// Since the grammar allows for both omitting the type and omitting the port
+/// name (the latter only if the subroutine has no body, i.e. is a prototype),
+/// there is an ambiguity when just a name is provided: `foo` may be a port with
+/// implicit type, if `foo` is not a defined type or the subroutine is not a
+/// prototype, or `foo` may be a port of type `foo` without a name, if `foo` is
+/// a defined type.
 #[moore_derive::node]
 #[indefinite("subroutine port")]
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct SubroutinePort<'a> {
+    /// The direction of the port, if provided.
     pub dir: Option<SubroutinePortDir>,
+    /// Whether an explicit `var` keyword was present.
     pub var: bool,
-    pub ty: Type<'a>,
-    pub name: Option<SubroutinePortName<'a>>,
+    /// The type and/or name of the port.
+    pub ty_name: Ambiguous<(Type<'a>, Option<SubroutinePortName<'a>>)>,
 }
 
+/// A single name of a function or task port.
+///
+/// This is the `x[3]` part in `inout logic [7:0] x[3]`.
 #[moore_derive::visit]
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct SubroutinePortName<'a> {
+    /// The name of the port.
     pub name: Spanned<Name>,
+    /// The unpacked dimensions of the port.
     pub dims: Vec<TypeDim<'a>>,
+    /// The optional default value of the port.
     pub expr: Option<Expr<'a>>,
 }
 
@@ -1742,13 +1800,18 @@ pub enum SubroutineItem<'a> {
     Stmt(Stmt<'a>),
 }
 
-#[moore_derive::visit]
+/// A function or task port declaration located in the body.
+#[moore_derive::node]
+#[indefinite("subroutine port")]
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct SubroutinePortDecl<'a> {
-    pub span: Span,
+    /// The direction of the port.
     pub dir: SubroutinePortDir,
+    /// Whether an explicit `var` keyword was present.
     pub var: bool,
+    /// The port type.
     pub ty: Type<'a>,
+    /// The port name declarations.
     pub names: Vec<VarDeclName<'a>>,
 }
 
@@ -1762,6 +1825,18 @@ pub enum SubroutinePortDir {
     ConstRef,
 }
 
+impl std::fmt::Display for SubroutinePortDir {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match self {
+            Self::Input => write!(f, "input"),
+            Self::Output => write!(f, "output"),
+            Self::Inout => write!(f, "inout"),
+            Self::Ref => write!(f, "ref"),
+            Self::ConstRef => write!(f, "const ref"),
+        }
+    }
+}
+
 /// A net declaration.
 ///
 /// For example `wire x, y, z`.
